#!/usr/bin/env python
# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
It defines the method with all required parameters to run restapi locally.
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import argparse
import imp
import os

from flask_cors import CORS

from aiida.backends.utils import load_dbenv


# pylint: disable=inconsistent-return-statements,too-many-locals
def run_api(flask_app, flask_api, *args, **kwargs):
    """
    Takes a flask.Flask instance and runs it. Parses
    command-line flags to configure the app.

    flask_app: Class inheriting from Flask app class
    flask_api = flask_restful API class to be used to wrap the app

    args: required by argparse

    kwargs:
    List of valid parameters:
    prog_name: name of the command before arguments are parsed. Useful when
    api is embedded in a command, such as verdi restapi
    default_host: self-explainatory
    default_port: self-explainatory
    default_config_dir = directory containing the config.py file used to
    configure the RESTapi
    parse_aiida_profile= if True, parses an option to specify the AiiDA
    profile
    All other passed parameters are ignored.
    """

    import aiida  # Mainly needed to locate the correct aiida path

    # Unpack parameters and assign defaults if needed
    prog_name = kwargs['prog_name'] if 'prog_name' in kwargs else ""

    default_host = kwargs['default_host'] if 'default_host' in kwargs else \
        "127.0.0.1"

    default_port = kwargs['default_port'] if 'default_port' in kwargs else \
        "5000"

    default_config_dir = kwargs['default_config_dir'] if \
        'default_config_dir' in kwargs \
        else os.path.join(os.path.split(os.path.abspath(
        aiida.restapi.__file__))[0], 'common')

    parse_aiida_profile = kwargs['parse_aiida_profile'] if \
        'parse_aiida_profile' in kwargs else False

    catch_internal_server = kwargs['catch_internal_server'] if\
        'catch_internal_server' in kwargs else False

    hookup = kwargs['hookup'] if 'hookup' in kwargs else False

    # Set up the command-line options
    parser = argparse.ArgumentParser(prog=prog_name, description='Hook up the AiiDA ' 'RESTful API')

    parser.add_argument("-H", "--host",
                        help="Hostname of the Flask app " + \
                             "[default %s]" % default_host,
                        dest='host',
                        default=default_host)
    parser.add_argument("-P", "--port",
                        help="Port for the Flask app " + \
                             "[default %s]" % default_port,
                        dest='port',
                        default=default_port)
    parser.add_argument("-c", "--config-dir",
                        help="Directory with config.py for Flask app " + \
                             "[default {}]".format(default_config_dir),
                        dest='config_dir',
                        default=default_config_dir)

    # This one is included only if necessary
    if parse_aiida_profile:
        parser.add_argument(
            "-p",
            "--aiida-profile",
            help="AiiDA profile to expose through the RESTful "
            "API [default: the default AiiDA profile]",
            dest="aiida_profile",
            default=None)

    # Two options useful for debugging purposes, but
    # a bit dangerous so not exposed in the help message.
    parser.add_argument("-d", "--debug", action="store_true", dest="debug", help=argparse.SUPPRESS)
    parser.add_argument("-w", "--wsgi-profile", action="store_true", dest="wsgi_profile", help=argparse.SUPPRESS)

    parsed_args = parser.parse_args(args)

    # Import the right configuration file
    confs = imp.load_source(
        os.path.join(parsed_args.config_dir, 'config'), os.path.join(parsed_args.config_dir, 'config.py'))

    import aiida.backends.settings as settings

    # Set aiida profile
    #
    # General logic:
    #
    # if aiida_profile is parsed the following cases exist:
    #
    # aiida_profile:
    #    "default"    --> default profile set in .aiida/config.json
    #    <profile>    --> corresponding profile in .aiida/config.json
    #    None         --> default restapi profile set in <config_dir>/config,py
    #
    # if aiida_profile is not parsed we assume
    #
    # default restapi profile set in <config_dir>/config.py

    if parse_aiida_profile and parsed_args.aiida_profile is not None:
        aiida_profile = parsed_args.aiida_profile

    elif confs.DEFAULT_AIIDA_PROFILE is not None:
        aiida_profile = confs.DEFAULT_AIIDA_PROFILE

    else:
        aiida_profile = "default"

    if aiida_profile != "default":
        settings.AIIDADB_PROFILE = aiida_profile
    else:
        pass  # This way the default of .aiida/config.json will be used

    # Set the AiiDA environment. If already loaded, load_dbenv will raise an
    # exception
    # if not is_dbenv_loaded():
    load_dbenv()

    # Instantiate an app
    app_kwargs = dict(catch_internal_server=catch_internal_server)
    app = flask_app(__name__, **app_kwargs)

    # Config the app
    app.config.update(**confs.APP_CONFIG)

    # cors
    cors_prefix = os.path.join(confs.PREFIX, "*")
    CORS(app, resources={r"" + cors_prefix: {"origins": "*"}})

    # Config the serializer used by the app
    if confs.SERIALIZER_CONFIG:
        from aiida.restapi.common.utils import CustomJSONEncoder
        app.json_encoder = CustomJSONEncoder

    # If the user selects the profiling option, then we need
    # to do a little extra setup
    if parsed_args.wsgi_profile:
        from werkzeug.contrib.profiler import ProfilerMiddleware

        app.config['PROFILE'] = True
        app.wsgi_app = ProfilerMiddleware(app.wsgi_app, restrictions=[30])

    # Instantiate an Api by associating its app
    api_kwargs = dict(PREFIX=confs.PREFIX, PERPAGE_DEFAULT=confs.PERPAGE_DEFAULT, LIMIT_DEFAULT=confs.LIMIT_DEFAULT)
    api = flask_api(app, **api_kwargs)

    # Check if the app has to be hooked-up or just returned
    if hookup:
        api.app.run(debug=parsed_args.debug, host=parsed_args.host, port=int(parsed_args.port), threaded=True)

    else:
        # here we return the app, and the api with no specifications on debug
        #  mode, port and host. This can be handled by an external server,
        # e.g. apache2, which will set the host and port. This implies that
        # the user-defined configuration of the app is ineffective (it only
        # affects the internal werkzeug server used by Flask).
        return (app, api)


# Standard boilerplate to run the api
if __name__ == '__main__':

    # I run the app via a wrapper that accepts arguments such as host and port
    # e.g. python api.py --host=127.0.0.2 --port=6000 --config-dir=~/.restapi
    # Default address is 127.0.0.1:5000, default config directory is
    # <aiida_path>/aiida/restapi/common
    #
    # Start the app by sliding the argvs to flaskrun, choose to take as an
    # argument also whether to parse the aiida profile or not (in verdi
    # restapi this would not be the case)

    import sys
    from aiida.restapi.api import AiidaApi, App

    #Or, equivalently, (useful starting point for derived apps)
    #import the app object and the Api class that you want to combine.

    run_api(App, AiidaApi, *sys.argv[1:], parse_aiida_profile=True, hookup=True, catch_internal_server=True)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""The main `verdi` click group."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from aiida.cmdline.params import options
from aiida.common.extendeddicts import AttributeDict
from aiida.common import exceptions


@click.group(invoke_without_command=True)
@options.PROFILE()
@click.option('--version', is_flag=True, is_eager=True, help='Print the version of AiiDA that is currently installed.')
@click.pass_context
def verdi(ctx, profile, version):
    """The command line interface of AiiDA."""
    import sys
    from aiida import get_version
    from aiida.backends import settings
    from aiida.cmdline.utils import echo
    from aiida.manage.configuration import get_config

    if version:
        echo.echo('AiiDA version {}'.format(get_version()))
        sys.exit(0)

    if ctx.obj is None:
        ctx.obj = AttributeDict()

    try:
        config = get_config()
    except exceptions.ConfigurationError:
        config = None
    else:

        if not profile:
            try:
                profile = config.get_profile()
            except exceptions.ProfileConfigurationError:
                profile = None

    if profile:
        settings.AIIDADB_PROFILE = profile.name

    ctx.obj.config = config
    ctx.obj.profile = profile

    ctx.help_option_names = ['-h', '--help']



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import os
from aiida.backends import settings
from aiida.common.exceptions import ConfigurationError, MissingConfigurationError
from aiida.common.setup import parse_repository_uri
from aiida.manage.configuration import get_config


TESTING_MODE = False

try:
    config = get_config()
except MissingConfigurationError:
    raise MissingConfigurationError("Please run the AiiDA installation, no config found")

if settings.AIIDADB_PROFILE is None:
    raise ConfigurationError("AIIDADB_PROFILE not defined, did you load django "
                             "through the AiiDA load_dbenv()?")

PROFILE = config.current_profile
PROFILE_CONF = PROFILE.dictionary

# put all database specific portions of settings here
BACKEND = PROFILE_CONF.get('AIIDADB_BACKEND', 'django')
DBENGINE = PROFILE_CONF.get('AIIDADB_ENGINE', '')
DBNAME = PROFILE_CONF.get('AIIDADB_NAME', '')
DBUSER = PROFILE_CONF.get('AIIDADB_USER', '')
DBPASS = PROFILE_CONF.get('AIIDADB_PASS', '')
DBHOST = PROFILE_CONF.get('AIIDADB_HOST', '')
DBPORT = PROFILE_CONF.get('AIIDADB_PORT', '')
REPOSITORY_URI = PROFILE_CONF.get('AIIDADB_REPOSITORY_URI', '')


# Checks on the REPOSITORY_* variables
try:
    REPOSITORY_URI
except NameError:
    raise ConfigurationError(
        "Please setup correctly the REPOSITORY_URI variable to "
        "a suitable directory on which you have write permissions.")

# Note: this variable might disappear in the future
REPOSITORY_PROTOCOL, REPOSITORY_PATH = parse_repository_uri(REPOSITORY_URI)

if settings.IN_RT_DOC_MODE:
    pass
elif REPOSITORY_PROTOCOL == 'file':
    if not os.path.isdir(REPOSITORY_PATH):
        try:
            # Try to create the local repository folders with needed parent
            # folders
            os.makedirs(REPOSITORY_PATH)
        except OSError:
            # Possibly here due to permission problems
            raise ConfigurationError(
                "Please setup correctly the REPOSITORY_PATH variable to "
                "a suitable directory on which you have write permissions. "
                "(I was not able to create the directory.)")
else:
    raise ConfigurationError("Only file protocol supported")



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.orm import KpointsData, Dict

__all__ = ('check_seekpath_is_installed', 'get_explicit_kpoints_path', 'get_kpoints_path')


def check_seekpath_is_installed():
    """
    Tries to import the Seekpath module. Raise ImportError if it cannot be imported

    :raises: ImportError
    """
    try:
        import seekpath
    except ImportError:
        raise ImportError("Seekpath is not installed, please install with 'pip install seekpath'")


def get_explicit_kpoints_path(structure, parameters):
    """
    Return the kpoint path for band structure (in scaled and absolute
    coordinates), given a crystal structure,
    using the paths proposed in the various publications (see description
    of the 'recipe' input parameter). The parameters are the same
    as get get_explicit_k_path in __init__, but here all structures are
    input and returned as AiiDA structures rather than tuples, and similarly
    k-points-related information as a AiiDA KpointsData class.

    :param structure: The AiiDA StructureData for which we want to obtain
        the suggested path.

    :param parameters: A dictionary whose key-value pairs are passed as
        additional kwargs to the ``seekpath.get_explicit_k_path`` function.

    :return: A dictionary with four nodes:

        - ``explicit_kpoints``: a KpointsData with the (explicit) kpoints
          (with labels set).

        - ``parameters``: a Dict, whose content is
          the same dictionary as returned by the ``seekpath.get_explicit_k_path`` function
          (see `seekpath documentation <https://seekpath.readthedocs.io/>`_),
          except that:

          - ``conv_lattice``, ``conv_positions``, ``conv_types``
            are removed and replaced by the ``conv_structure`` output node

          - ``primitive_lattice``, ``primitive_positions``, ``primitive_types``
            are removed and replaced by the `primitive_structure` output node

          - ``reciprocal_primitive_lattice``, ``explicit_kpoints_abs``,
            ``explicit_kpoints_rel`` and ``explicit_kpoints_labels`` are removed
            and replaced by the ``explicit_kpoints`` output node

        - ``primitive_structure``: A StructureData with the primitive structure

        - ``conv_structure``: A StructureData with the primitive structure
    """
    from aiida.tools.data.structure import spglib_tuple_to_structure, structure_to_spglib_tuple

    check_seekpath_is_installed()
    import seekpath

    structure_tuple, kind_info, kinds = structure_to_spglib_tuple(structure)

    result = {}
    rawdict = seekpath.get_explicit_k_path(structure=structure_tuple, **parameters)

    # Replace primitive structure with AiiDA StructureData
    primitive_lattice = rawdict.pop('primitive_lattice')
    primitive_positions = rawdict.pop('primitive_positions')
    primitive_types = rawdict.pop('primitive_types')
    primitive_tuple = (primitive_lattice, primitive_positions, primitive_types)
    primitive_structure = spglib_tuple_to_structure(primitive_tuple, kind_info, kinds)

    # Replace conv structure with AiiDA StructureData
    conv_lattice = rawdict.pop('conv_lattice')
    conv_positions = rawdict.pop('conv_positions')
    conv_types = rawdict.pop('conv_types')
    conv_tuple = (conv_lattice, conv_positions, conv_types)
    conv_structure = spglib_tuple_to_structure(conv_tuple, kind_info, kinds)

    # Remove reciprocal_primitive_lattice, recalculated by kpoints class
    rawdict.pop('reciprocal_primitive_lattice')
    kpoints_abs = rawdict.pop('explicit_kpoints_abs')
    kpoints_rel = rawdict.pop('explicit_kpoints_rel')
    kpoints_labels = rawdict.pop('explicit_kpoints_labels')

    # set_kpoints expects labels like [[0,'X'],[34,'L'],...], so generate it here skipping empty labels
    labels = [[idx, label] for idx, label in enumerate(kpoints_labels) if label]
    kpoints = KpointsData()
    kpoints.set_cell_from_structure(primitive_structure)
    kpoints.set_kpoints(kpoints_abs, cartesian=True, labels=labels)

    result['parameters'] = Dict(dict=rawdict)
    result['explicit_kpoints'] = kpoints
    result['primitive_structure'] = primitive_structure
    result['conv_structure'] = conv_structure

    return result


def get_kpoints_path(structure, parameters):
    """
    Return the kpoint path information for band structure given a
    crystal structure, using the paths from the chosen recipe/reference.
    The parameters are the same
    as get get_path in __init__, but here all structures are
    input and returned as AiiDA structures rather than tuples.


    If you use this module, please cite the paper of the corresponding
    recipe (see documentation of seekpath).

    :param structure: The crystal structure for which we want to obtain
        the suggested path. It should be an AiiDA StructureData object.

    :param parameters: A dictionary whose key-value pairs are passed as
        additional kwargs to the ``seekpath.get_path`` function.

    :return: A dictionary with three nodes:

        - ``parameters``: a Dict, whose content is
          the same dictionary as returned by the ``seekpath.get_path`` function
          (see `seekpath documentation <https://seekpath.readthedocs.io/>`_),
          except that:

          - ``conv_lattice``, ``conv_positions``, ``conv_types``
            are removed and replaced by the ``conv_structure`` output node

          - ``primitive_lattice``, ``primitive_positions``, ``primitive_types``
            are removed and replaced by the ``primitive_structure`` output node

        - ``primitive_structure``: A StructureData with the primitive structure

        - ``conv_structure``: A StructureData with the primitive structure
    """
    from aiida.tools.data.structure import spglib_tuple_to_structure, structure_to_spglib_tuple

    check_seekpath_is_installed()
    import seekpath

    structure_tuple, kind_info, kinds = structure_to_spglib_tuple(structure)

    result = {}
    rawdict = seekpath.get_path(structure=structure_tuple, **parameters)

    # Replace conv structure with AiiDA StructureData
    conv_lattice = rawdict.pop('conv_lattice')
    conv_positions = rawdict.pop('conv_positions')
    conv_types = rawdict.pop('conv_types')
    conv_tuple = (conv_lattice, conv_positions, conv_types)
    conv_structure = spglib_tuple_to_structure(conv_tuple, kind_info, kinds)

    # Replace primitive structure with AiiDA StructureData
    primitive_lattice = rawdict.pop('primitive_lattice')
    primitive_positions = rawdict.pop('primitive_positions')
    primitive_types = rawdict.pop('primitive_types')
    primitive_tuple = (primitive_lattice, primitive_positions, primitive_types)
    primitive_structure = spglib_tuple_to_structure(primitive_tuple, kind_info, kinds)

    result['parameters'] = Dict(dict=rawdict)
    result['primitive_structure'] = primitive_structure
    result['conv_structure'] = conv_structure

    return result



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module with `OrmEntityLoader` and its sub classes that simplify loading entities through their identifiers."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from abc import ABCMeta
from enum import Enum

import six

from aiida.common.exceptions import MultipleObjectsError, NotExistent
from aiida.common.lang import abstractclassmethod, classproperty
from aiida.orm.querybuilder import QueryBuilder

__all__ = ('get_loader', 'OrmEntityLoader', 'CalculationEntityLoader', 'CodeEntityLoader', 'ComputerEntityLoader',
           'GroupEntityLoader', 'NodeEntityLoader')


def get_loader(orm_class):
    """Return the correct OrmEntityLoader for the given orm class.

    :param orm_class: the orm class
    :returns: a subclass of OrmEntityLoader
    :raises ValueError: if no OrmEntityLoader subclass can be found for the given orm class
    """
    from aiida.orm import Code, Computer, Group, Node

    if issubclass(orm_class, Code):
        return CodeEntityLoader

    if issubclass(orm_class, Computer):
        return ComputerEntityLoader

    if issubclass(orm_class, Group):
        return GroupEntityLoader

    if issubclass(orm_class, Node):
        return NodeEntityLoader

    raise ValueError('no OrmEntityLoader available for {}'.format(orm_class))


class IdentifierType(Enum):
    """
    The enumeration that defines the three types of identifier that can be used to identify an orm entity.
    The ID is always an integer, the UUID a base 16 encoded integer with optional dashes and the LABEL can
    be any string based label or name, the format of which will vary per orm class
    """
    # pylint: disable=invalid-name

    ID = 'ID'
    UUID = 'UUID'
    LABEL = 'LABEL'


@six.add_metaclass(ABCMeta)
class OrmEntityLoader(object):
    """Base class for entity loaders."""

    # pylint: disable=useless-object-inheritance

    label_ambiguity_breaker = '!'

    @classproperty
    def orm_base_class(self):
        """
        Return the orm base class to which loaded entities should be mapped. Actual queries to load an entity
        may further narrow the query set by defining a more specific set of orm classes, as long as each of
        those is a strict sub class of the orm base class.

        :returns: the orm base class
        """
        raise NotImplementedError

    @abstractclassmethod
    def _get_query_builder_label_identifier(cls, identifier, classes):
        """
        Return the query builder instance that attempts to map the identifier onto an entity of the orm class,
        defined for this loader class, interpreting the identifier as a LABEL like identifier

        :param identifier: the LABEL identifier
        :param classes: a tuple of orm classes to which the identifier should be mapped
        :returns: the query builder instance
        :raises ValueError: if the identifier is invalid
        :raises aiida.common.NotExistent: if the orm base class does not support a LABEL like identifier
        """
        raise NotImplementedError

    @classmethod
    def _get_query_builder_id_identifier(cls, identifier, classes):
        """
        Return the query builder instance that attempts to map the identifier onto an entity of the orm class,
        defined for this loader class, interpreting the identifier as an ID like identifier

        :param identifier: the ID identifier
        :param classes: a tuple of orm classes to which the identifier should be mapped
        :returns: the query builder instance
        """
        builder = QueryBuilder()
        builder.append(cls=classes, tag='entity', project=['*'])
        builder.add_filter('entity', {'id': identifier})

        return builder

    @classmethod
    def _get_query_builder_uuid_identifier(cls, identifier, classes, query_with_dashes):
        """
        Return the query builder instance that attempts to map the identifier onto an entity of the orm class,
        defined for this loader class, interpreting the identifier as a UUID like identifier

        :param identifier: the UUID identifier
        :param classes: a tuple of orm classes to which the identifier should be mapped
        :returns: the query builder instance
        """
        from uuid import UUID

        uuid = identifier.replace('-', '')

        if query_with_dashes:
            for dash_pos in [20, 16, 12, 8]:
                if len(uuid) > dash_pos:
                    uuid = '{}-{}'.format(uuid[:dash_pos], uuid[dash_pos:])

        builder = QueryBuilder()
        builder.append(cls=classes, tag='entity', project=['*'])

        # If a UUID can be constructed from the identifier, it is a full UUID and the query can use an equality operator
        try:
            UUID(uuid)
        except ValueError:
            builder.add_filter('entity', {'uuid': {'like': '{}%'.format(uuid)}})
        else:
            builder.add_filter('entity', {'uuid': uuid})

        return builder

    @classmethod
    def get_query_builder(cls, identifier, identifier_type=None, sub_classes=None, query_with_dashes=True):
        """
        Return the query builder instance that attempts to map the identifier onto an entity of the orm class,
        defined for this loader class, inferring the identifier type if it is not defined.

        :param identifier: the identifier
        :param identifier_type: the type of the identifier
        :param sub_classes: an optional tuple of orm classes, that should each be strict sub classes of the
            base orm class of the loader, that will narrow the queryset
        :returns: the query builder instance and a dictionary of used query parameters
        """
        classes = cls.get_query_classes(sub_classes)

        if identifier_type is None:
            identifier, identifier_type = cls.infer_identifier_type(identifier)

        if identifier_type == IdentifierType.ID:
            builder = cls._get_query_builder_id_identifier(identifier, classes)
        elif identifier_type == IdentifierType.UUID:
            builder = cls._get_query_builder_uuid_identifier(identifier, classes, query_with_dashes)
        elif identifier_type == IdentifierType.LABEL:
            builder = cls._get_query_builder_label_identifier(identifier, classes)

        query_parameters = {
            'classes': classes,
            'identifier': identifier,
            'identifier_type': identifier_type,
        }

        return builder, query_parameters

    @classmethod
    def load_entity(cls, identifier, identifier_type=None, sub_classes=None, query_with_dashes=True):
        """
        Load an entity that uniquely corresponds to the provided identifier of the identifier type.

        :param identifier: the identifier
        :param identifier_type: the type of the identifier
        :param sub_classes: an optional tuple of orm classes, that should each be strict sub classes of the
            base orm class of the loader, that will narrow the queryset
        :returns: the loaded entity
        :raises aiida.common.MultipleObjectsError: if the identifier maps onto multiple entities
        :raises aiida.common.NotExistent: if the identifier maps onto not a single entity
        """
        builder, query_parameters = cls.get_query_builder(identifier, identifier_type, sub_classes, query_with_dashes)
        builder.limit(2)

        classes = ' or '.join([sub_class.__name__ for sub_class in query_parameters['classes']])
        identifier = query_parameters['identifier']
        identifier_type = query_parameters['identifier_type'].value

        try:
            entity = builder.one()[0]
        except MultipleObjectsError:
            error = 'multiple {} entries found with {}<{}>'.format(classes, identifier_type, identifier)
            raise MultipleObjectsError(error)
        except NotExistent as exception:
            error = 'no {} found with {}<{}>: {}'.format(classes, identifier_type, identifier, exception)
            raise NotExistent(error)

        return entity

    @classmethod
    def get_query_classes(cls, sub_classes=None):
        """
        Get the tuple of classes to be used for the entity query. If sub_classes is defined, each class will be
        validated by verifying that it is a sub class of the loader's orm base class.
        Validate a tuple of classes if a user passes in a specific one when attempting to load an entity. Each class
        should be a sub class of the entity loader's orm base class

        :param sub_classes: an optional tuple of orm classes, that should each be strict sub classes of the
            base orm class of the loader, that will narrow the queryset
        :returns: the tuple of orm classes to be used for the entity query
        :raises ValueError: if any of the classes are not a sub class of the entity loader's orm base class
        """
        if sub_classes is None:
            return (cls.orm_base_class,)

        if not isinstance(sub_classes, tuple):
            raise TypeError('sub_classes should be a tuple: {}'.format(sub_classes))

        for sub_class in sub_classes:
            if not issubclass(sub_class, cls.orm_base_class):
                raise ValueError('{} is not a sub class of the base orm class {}'.format(sub_class, cls.orm_base_class))

        return sub_classes

    @classmethod
    def infer_identifier_type(cls, value):
        """
        This method will attempt to automatically distinguish which identifier type is implied for the given value, if
        the value itself has no type from which it can be inferred.

        The strategy is to first attempt to convert the value to an integer. If successful, it is assumed that the value
        represents an ID. If that fails, we attempt to interpret the value as a base 16 encoded integer, after having
        removed any dashes from the string. If that succeeds, it is most likely a UUID. If it seems to be neither an ID
        nor a UUID, it is assumed to be a LABEL like identifier.

        With this approach there is the possibility for ambiguity. Since it is allowed to pass a partial UUID, it is
        possible that the partial UUID is also a valid ID. Likewise, a LABEL identifier might also be a valid ID, or a
        valid (partial) UUID. Fortunately, these ambiguities can be solved though:

         * ID/UUID: can always be solved by passing a partial UUID with at least one dash
         * ID/LABEL: appending an exclamation point ! to the identifier, will force LABEL interpretation
         * UUID/LABEL: appending an exclamation point ! to the identifier, will force LABEL interpretation

        As one can see, the user will always be able to include at least one dash of the UUID identifier to prevent it
        from being interpreted as an ID. For the potential ambiguities in LABEL identifiers, we had to introduce a
        special marker to provide a surefire way of breaking any ambiguity that may arise. Adding an exclamation point
        will break the normal strategy and the identifier will directly be interpreted as a LABEL identifier.

        :param value: the value of the identifier
        :returns: the identifier and identifier type
        :raises ValueError: if the value is an invalid identifier
        """
        if not value:
            raise ValueError('the value for the identifier cannot be empty')

        if not isinstance(value, six.string_types):
            value = str(value)

        # If the final character of the value is the special marker, we enforce LABEL interpretation
        if value[-1] == cls.label_ambiguity_breaker:

            identifier = value.rstrip(cls.label_ambiguity_breaker)
            identifier_type = IdentifierType.LABEL

        else:

            # If the value can be cast into an integer, interpret it as an ID
            try:
                identifier = int(value)
                identifier_type = IdentifierType.ID
            except ValueError:

                # If the value is a valid base sixteen encoded integer, after dashes are removed, interpret it as a UUID
                try:
                    int(value.replace('-', ''), 16)
                    identifier = value
                    identifier_type = IdentifierType.UUID

                # If that fails, interpret it as a LABEL
                except ValueError:
                    identifier = value
                    identifier_type = IdentifierType.LABEL

        return identifier, identifier_type


class ProcessEntityLoader(OrmEntityLoader):
    """Loader for the `Process` entity and sub classes."""

    @classproperty
    def orm_base_class(self):
        """
        Return the orm base class to which loaded entities should be mapped. Actual queries to load an entity
        may further narrow the query set by defining a more specific set of orm classes, as long as each of
        those is a strict sub class of the orm base class.

        :returns: the orm base class
        """
        from aiida.orm import ProcessNode
        return ProcessNode

    @classmethod
    def _get_query_builder_label_identifier(cls, identifier, classes):
        """
        Return the query builder instance that attempts to map the identifier onto an entity of the orm class,
        defined for this loader class, interpreting the identifier as a LABEL like identifier

        :param identifier: the LABEL identifier
        :param classes: a tuple of orm classes to which the identifier should be mapped
        :returns: the query builder instance that should retrieve the entity corresponding to the identifier
        :raises ValueError: if the identifier is invalid
        :raises aiida.common.NotExistent: if the orm base class does not support a LABEL like identifier
        """
        builder = QueryBuilder()
        builder.append(cls=classes, tag='process', project=['*'], filters={'label': {'==': identifier}})

        return builder


class CalculationEntityLoader(OrmEntityLoader):
    """Loader for the `Calculation` entity and sub classes."""

    @classproperty
    def orm_base_class(self):
        """
        Return the orm base class to which loaded entities should be mapped. Actual queries to load an entity
        may further narrow the query set by defining a more specific set of orm classes, as long as each of
        those is a strict sub class of the orm base class.

        :returns: the orm base class
        """
        from aiida.orm import CalculationNode
        return CalculationNode

    @classmethod
    def _get_query_builder_label_identifier(cls, identifier, classes):
        """
        Return the query builder instance that attempts to map the identifier onto an entity of the orm class,
        defined for this loader class, interpreting the identifier as a LABEL like identifier

        :param identifier: the LABEL identifier
        :param classes: a tuple of orm classes to which the identifier should be mapped
        :returns: the query builder instance that should retrieve the entity corresponding to the identifier
        :raises ValueError: if the identifier is invalid
        :raises aiida.common.NotExistent: if the orm base class does not support a LABEL like identifier
        """
        builder = QueryBuilder()
        builder.append(cls=classes, tag='calculation', project=['*'], filters={'label': {'==': identifier}})

        return builder


class WorkflowEntityLoader(OrmEntityLoader):
    """Loader for the `Workflow` entity and sub classes."""

    @classproperty
    def orm_base_class(self):
        """
        Return the orm base class to which loaded entities should be mapped. Actual queries to load an entity
        may further narrow the query set by defining a more specific set of orm classes, as long as each of
        those is a strict sub class of the orm base class.

        :returns: the orm base class
        """
        from aiida.orm import WorkflowNode
        return WorkflowNode

    @classmethod
    def _get_query_builder_label_identifier(cls, identifier, classes):
        """
        Return the query builder instance that attempts to map the identifier onto an entity of the orm class,
        defined for this loader class, interpreting the identifier as a LABEL like identifier

        :param identifier: the LABEL identifier
        :param classes: a tuple of orm classes to which the identifier should be mapped
        :returns: the query builder instance that should retrieve the entity corresponding to the identifier
        :raises ValueError: if the identifier is invalid
        :raises aiida.common.NotExistent: if the orm base class does not support a LABEL like identifier
        """
        builder = QueryBuilder()
        builder.append(cls=classes, tag='workflow', project=['*'], filters={'label': {'==': identifier}})

        return builder


class CodeEntityLoader(OrmEntityLoader):
    """Loader for the `Code` entity and sub classes."""

    @classproperty
    def orm_base_class(self):
        """
        Return the orm base class to which loaded entities should be mapped. Actual queries to load an entity
        may further narrow the query set by defining a more specific set of orm classes, as long as each of
        those is a strict sub class of the orm base class.

        :returns: the orm base class
        """
        from aiida.orm import Code
        return Code

    @classmethod
    def _get_query_builder_label_identifier(cls, identifier, classes):
        """
        Return the query builder instance that attempts to map the identifier onto an entity of the orm class,
        defined for this loader class, interpreting the identifier as a LABEL like identifier

        :param identifier: the LABEL identifier
        :param classes: a tuple of orm classes to which the identifier should be mapped
        :returns: the query builder instance that should retrieve the entity corresponding to the identifier
        :raises ValueError: if the identifier is invalid
        :raises aiida.common.NotExistent: if the orm base class does not support a LABEL like identifier
        """
        from aiida.orm import Computer

        try:
            label, _, machinename = identifier.partition('@')
        except AttributeError:
            raise ValueError('the identifier needs to be a string')

        builder = QueryBuilder()
        builder.append(cls=classes, tag='code', project=['*'], filters={'label': {'==': label}})

        if machinename:
            builder.append(Computer, filters={'name': {'==': machinename}}, with_node='code')

        return builder


class ComputerEntityLoader(OrmEntityLoader):
    """Loader for the `Computer` entity and sub classes."""

    @classproperty
    def orm_base_class(self):
        """
        Return the orm base class to which loaded entities should be mapped. Actual queries to load an entity
        may further narrow the query set by defining a more specific set of orm classes, as long as each of
        those is a strict sub class of the orm base class.

        :returns: the orm base class
        """
        from aiida.orm import Computer
        return Computer

    @classmethod
    def _get_query_builder_label_identifier(cls, identifier, classes):
        """
        Return the query builder instance that attempts to map the identifier onto an entity of the orm class,
        defined for this loader class, interpreting the identifier as a LABEL like identifier

        :param identifier: the LABEL identifier
        :param classes: a tuple of orm classes to which the identifier should be mapped
        :returns: the query builder instance that should retrieve the entity corresponding to the identifier
        :raises ValueError: if the identifier is invalid
        :raises aiida.common.NotExistent: if the orm base class does not support a LABEL like identifier
        """
        builder = QueryBuilder()
        builder.append(cls=classes, tag='computer', project=['*'], filters={'name': {'==': identifier}})

        return builder


class DataEntityLoader(OrmEntityLoader):
    """Loader for the `Data` entity and sub classes."""

    @classproperty
    def orm_base_class(self):
        """
        Return the orm base class to which loaded entities should be mapped. Actual queries to load an entity
        may further narrow the query set by defining a more specific set of orm classes, as long as each of
        those is a strict sub class of the orm base class.

        :returns: the orm base class
        """
        from aiida.orm import Data
        return Data

    @classmethod
    def _get_query_builder_label_identifier(cls, identifier, classes):
        """
        Return the query builder instance that attempts to map the identifier onto an entity of the orm class,
        defined for this loader class, interpreting the identifier as a LABEL like identifier

        :param identifier: the LABEL identifier
        :param classes: a tuple of orm classes to which the identifier should be mapped
        :returns: the query builder instance that should retrieve the entity corresponding to the identifier
        :raises ValueError: if the identifier is invalid
        :raises aiida.common.NotExistent: if the orm base class does not support a LABEL like identifier
        """
        builder = QueryBuilder()
        builder.append(cls=classes, tag='calculation', project=['*'], filters={'label': {'==': identifier}})

        return builder


class GroupEntityLoader(OrmEntityLoader):
    """Loader for the `Group` entity and sub classes."""

    @classproperty
    def orm_base_class(self):
        """
        Return the orm base class to which loaded entities should be mapped. Actual queries to load an entity
        may further narrow the query set by defining a more specific set of orm classes, as long as each of
        those is a strict sub class of the orm base class.

        :returns: the orm base class
        """
        from aiida.orm.groups import Group
        return Group

    @classmethod
    def _get_query_builder_label_identifier(cls, identifier, classes):
        """
        Return the query builder instance that attempts to map the identifier onto an entity of the orm class,
        defined for this loader class, interpreting the identifier as a LABEL like identifier

        :param identifier: the LABEL identifier
        :param classes: a tuple of orm classes to which the identifier should be mapped
        :returns: the query builder instance that should retrieve the entity corresponding to the identifier
        :raises ValueError: if the identifier is invalid
        :raises aiida.common.NotExistent: if the orm base class does not support a LABEL like identifier
        """
        builder = QueryBuilder()
        builder.append(cls=classes, tag='group', project=['*'], filters={'label': {'==': identifier}})

        return builder


class NodeEntityLoader(OrmEntityLoader):
    """Loader for the `Node` entity and sub classes."""

    @classproperty
    def orm_base_class(self):
        """
        Return the orm base class to which loaded entities should be mapped. Actual queries to load an entity
        may further narrow the query set by defining a more specific set of orm classes, as long as each of
        those is a strict sub class of the orm base class.

        :returns: the orm base class
        """
        from aiida.orm import Node
        return Node

    @classmethod
    def _get_query_builder_label_identifier(cls, identifier, classes):
        """
        Return the query builder instance that attempts to map the identifier onto an entity of the orm class,
        defined for this loader class, interpreting the identifier as a LABEL like identifier

        :param identifier: the LABEL identifier
        :param classes: a tuple of orm classes to which the identifier should be mapped
        :returns: the query builder instance that should retrieve the entity corresponding to the identifier
        :raises ValueError: if the identifier is invalid
        :raises aiida.common.NotExistent: if the orm base class does not support a LABEL like identifier
        """
        builder = QueryBuilder()
        builder.append(cls=classes, tag='node', project=['*'], filters={'label': {'==': identifier}})

        return builder



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Draw the provenance graphs
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import io
import os
import subprocess
import tempfile


def draw_graph(origin_node,
               ancestor_depth=None,
               descendant_depth=None,
               image_format='dot',
               include_calculation_inputs=False,
               include_calculation_outputs=False):
    """
    The algorithm starts from the original node and goes both input-ward and output-ward via a breadth-first algorithm.

    :param origin_node: An Aiida node, the starting point for drawing the graph
    :param int ancestor_depth: The maximum depth of the ancestors drawn. If left to None, we recurse until the graph is
        fully explored
    :param int descendant_depth: The maximum depth of the descendants drawn. If left to None, we recurse until the graph
        is fully explored
    :param str image_format: The output plot format, by default dot

    :returns: The exit_code of the subprocess.call() method that produced the valid file
    :returns: The file name of the final output

    ..note::
        If an invalid format is provided graphviz prints a helpful message, so this doesn't need to be implemented here.
    """
    # pylint: disable=too-many-locals,too-many-statements,too-many-branches
    from aiida.orm import ProcessNode
    from aiida.orm import Code
    from aiida.orm import Node
    from aiida.common.links import LinkType
    from aiida.orm.querybuilder import QueryBuilder

    def draw_node_settings(node, **kwargs):
        """
        Returns a string with all infos needed in a .dot file  to define a node of a graph.
        :param node:
        :param kwargs: Additional key-value pairs to be added to the returned string
        :return: a string
        """
        if isinstance(node, ProcessNode):
            shape = "shape=polygon,sides=4"
        elif isinstance(node, Code):
            shape = "shape=diamond"
        else:
            shape = "shape=ellipse"
        if kwargs:
            additional_params = ",{}".format(",".join('{}="{}"'.format(k, v) for k, v in kwargs.items()))
        else:
            additional_params = ""
        if node.label:
            label_string = "\\n'{}'".format(node.label)
            additional_string = ""
        else:
            additional_string = "\\n {}".format(node.get_description())
            label_string = ""
        labelstring = 'label="{} ({}){}{}"'.format(node.__class__.__name__, node.pk, label_string, additional_string)
        return "N{} [{},{}{}];".format(node.pk, shape, labelstring, additional_params)

    def draw_link_settings(inp_id, out_id, link_label, link_type):
        """Return a string with label information."""
        if link_type in (LinkType.CREATE.value, LinkType.INPUT_CALC.value, LinkType.INPUT_WORK.value):
            style = 'solid'  # Solid lines and black colors
            color = "0.0 0.0 0.0"  # for CREATE and INPUT (The provenance graph)
        elif link_type == LinkType.RETURN.value:
            style = 'dotted'  # Dotted  lines of
            color = "0.0 0.0 0.0"  # black color for Returns
        elif link_type == (LinkType.CALL_CALC.value or LinkType.CALL_WORK.value):
            style = 'bold'  # Bold lines and
            color = "0.0 1.0 1.0"  # Bright red for calls
        else:
            style = 'solid'  # Solid and
            color = "0.0 0.0 0.5"  #grey lines for unspecified links!
        return '    {} -> {} [label="{}", color="{}", style="{}"];'.format("N{}".format(inp_id), "N{}".format(out_id),
                                                                           link_label, color, style)

    # Breadth-first search of all ancestors and descendant nodes of a given node
    links = {}  # Accumulate links here
    nodes = {
        origin_node.pk: draw_node_settings(origin_node, style='filled', color='lightblue')
    }  #Accumulate nodes specs here
    # Additional nodes (the ones added with either one of  include_calculation_inputs or include_calculation_outputs
    # is set to true. I have to put them in a different dictionary because nodes is the one used for the recursion,
    # whereas these should not be used for the recursion:
    additional_nodes = {}

    last_nodes = [origin_node]  # Put the nodes whose links have not been scanned yet

    # Go through the graph on-ward (i.e. look at inputs)
    depth = 0
    while last_nodes:
        # I augment depth every time I get through a new iteration
        depth += 1
        # I check whether I should stop here:
        if ancestor_depth is not None and depth > ancestor_depth:
            break
        # I continue by adding new nodes here!
        new_nodes = []
        for node in last_nodes:
            # This query gives me all the inputs of this node, and link labels and types!
            input_query = QueryBuilder()
            input_query.append(Node, filters={'id': node.pk}, tag='n')
            input_query.append(Node, with_outgoing='n', edge_project=('id', 'label', 'type'), project='*', tag='inp')
            for inp, link_id, link_label, link_type in input_query.iterall():
                # I removed this check, to me there is no way that this link was already referred to!
                # if link_id not in links:
                links[link_id] = draw_link_settings(inp.pk, node.pk, link_label, link_type)
                # For the nodes I need to check, maybe this same node is referred to multiple times.
                if inp.pk not in nodes:
                    nodes[inp.pk] = draw_node_settings(inp)
                    new_nodes.append(inp)

            # Checking whether I also should include all the outputs of a calculation into the drawing:
            if include_calculation_outputs and isinstance(node, ProcessNode):
                # Query for the outputs, giving me also link labels and types:
                output_query = QueryBuilder()
                output_query.append(Node, filters={'id': node.pk}, tag='n')
                output_query.append(
                    Node, with_incoming='n', edge_project=('id', 'label', 'type'), project='*', tag='out')
                # Iterate through results
                for out, link_id, link_label, link_type in output_query.iterall():
                    # This link might have been drawn already, because the output is maybe
                    # already drawn.
                    # To check: Maybe it's more efficient not to check this, since
                    # the dictionaries are large and contain many keys...
                    # I.e. just always draw, also when overwriting an existing (identical) entry.
                    if link_id not in links:
                        links[link_id] = draw_link_settings(node.pk, out.pk, link_label, link_type)
                    if out.pk not in nodes and out.pk not in additional_nodes:
                        additional_nodes[out.pk] = draw_node_settings(out)

        last_nodes = new_nodes

    # Go through the graph down-ward (i.e. look at outputs)
    last_nodes = [origin_node]
    depth = 0
    while last_nodes:
        depth += 1
        # Also here, checking of maximum descendant depth is set and applies.
        if descendant_depth is not None and depth > descendant_depth:
            break
        new_nodes = []

        for node in last_nodes:
            # Query for the outputs:
            output_query = QueryBuilder()
            output_query.append(Node, filters={'id': node.pk}, tag='n')
            output_query.append(Node, with_incoming='n', edge_project=('id', 'label', 'type'), project='*', tag='out')

            for out, link_id, link_label, link_type in output_query.iterall():
                # Draw the link
                links[link_id] = draw_link_settings(node.pk, out.pk, link_label, link_type)
                if out.pk not in nodes:
                    nodes[out.pk] = draw_node_settings(out)
                    new_nodes.append(out)

            if include_calculation_inputs and isinstance(node, ProcessNode):
                input_query = QueryBuilder()
                input_query.append(Node, filters={'id': node.pk}, tag='n')
                input_query.append(
                    Node, with_outgoing='n', edge_project=('id', 'label', 'type'), project='*', tag='inp')
                for inp, link_id, link_label, link_type in input_query.iterall():
                    # Also here, maybe it's just better not to check?
                    if link_id not in links:
                        links[link_id] = draw_link_settings(inp.pk, node.pk, link_label, link_type)
                    if inp.pk not in nodes and inp.pk not in additional_nodes:
                        additional_nodes[inp.pk] = draw_node_settings(inp)
        last_nodes = new_nodes

    # Writing the graph to a temporary file
    _, fname = tempfile.mkstemp(suffix='.dot')
    with io.open(fname, 'w', encoding='utf8') as fhandle:
        fhandle.write(u"digraph G {\n")
        for _, l_values in links.items():
            fhandle.write(u'    {}\n'.format(l_values))
        for _, n_values in nodes.items():
            fhandle.write(u"    {}\n".format(n_values))
        for _, n_values in additional_nodes.items():
            fhandle.write(u"    {}\n".format(n_values))
        fhandle.write(u"}\n")

    # Now I am producing the output file
    output_file_name = "{0}.{1}".format(origin_node.pk, image_format)
    # Try and convert the .dot file using the `dot` utility from graphviz
    try:
        exit_code = subprocess.call(['dot', '-T', image_format, fname, '-o', output_file_name])
    except OSError:
        from aiida.cmdline.utils import echo
        echo.echo_critical('Operating system error - perhaps Graphviz is not installed?')
    # cleaning up by removing the temporary file
    os.remove(fname)
    return exit_code, output_file_name
